"""
Copyright 2023 EMD International
License for this script: MIT https://opensource.org/license/mit/
License for windPRO commercial software: https://www.emd-international.com/contact-us/general-terms-conditions-sale/
"""

import os
import math
from windproapi.utils import get_windpro_sample_path
from windproapi import WindProApi
from windproapi import nan_to_skipvalue

_windproapi = WindProApi()
_windproapi.start_windpro_random_port()

# This folder needs to exist on your machine
working_dir = os.path.join(get_windpro_sample_path('4.0'), 'Solar')
os.makedirs(working_dir, exist_ok=True)
project_path = os.path.join(working_dir, 'Solar.w40p')

# Get to all services
online_data_service = _windproapi.get_service('OnlineDataService')
project_service = _windproapi.get_service('ProjectService')
objects_service = _windproapi.get_service('ObjectsService')
calculation_service = _windproapi.get_service('CalculationService')
calc_park_service = _windproapi.get_service('CalcParkService')
obj_meteo_service = _windproapi.get_service('ObjMeteoService')
obj_wtg_service = _windproapi.get_service('ObjWtgService')
obj_elevation_grid_service = _windproapi.get_service('ObjElevationGridService')
obj_solar_pv_service = _windproapi.get_service('ObjSolarPvService')
calc_solar_pv_service = _windproapi.get_service('CalcSolarPvService')
windpro_service = _windproapi.get_service('WindproService')
factory = _windproapi.get_factory('WindproService')

# Position of the site center.
lng = 13
lat = 56.8

# Size of the solar pv area
d_west_east = 500
d_south_north = 1000

# Paths where solar data is usually stored
path_pv_panels = os.path.join(os.path.dirname(get_windpro_sample_path('4.0')), "PVPanels")
path_pv_dea = os.path.join(os.path.dirname(get_windpro_sample_path('4.0')), "3D.dae_models", "Solar panels")

# Make a new project
project_service.NewProject(lng=lng, lat=lat, filename=project_path)

# Setting the project coordinate system to UTM
utm_epsg = project_service.GetUtmWgs84EpsgForProject()
project_service.SetProjectCoorEPSG(epsg=utm_epsg)

# UTM coordinates of center
utm_coords = windpro_service.ConvertCoorEPSG(x=lng,
                                             y=lat,
                                             inEpsg=4326,
                                             toEpsg=utm_epsg)
site_center_easting = utm_coords.X
site_center_northing = utm_coords.Y

# Adding a solar object
newObj = objects_service.AddObject('SolarPvObject', lat, lng, 'New SolarPvObject')
solar_obj = obj_solar_pv_service.GetSolarPvObject(newObj.Handle)

# Defining the area
def lat_lng_from_sitecenter(dx, dy):
    temp_coords = windpro_service.ConvertCoorEPSG(x=site_center_easting + dx,
                                                 y=site_center_northing + dy,
                                                 inEpsg=utm_epsg,
                                                 toEpsg=4326)
    return temp_coords.X, temp_coords.Y

solar_obj.SolarAreas = factory.TApiSolarAreas()
solar_area = factory.TApiSolarArea()
solar_area.Points = factory.TApiGPs()
newGP = factory.TApiGP()
newGP.Lng, newGP.Lat = lat_lng_from_sitecenter(d_west_east/2, d_south_north/2)
solar_area.Points.TApiGP.append(newGP)

newGP = factory.TApiGP()
newGP.Lng, newGP.Lat = lat_lng_from_sitecenter(-d_west_east/2, d_south_north/2)
solar_area.Points.TApiGP.append(newGP)

newGP = factory.TApiGP()
newGP.Lng, newGP.Lat = lat_lng_from_sitecenter(-d_west_east/2, -d_south_north/2)
solar_area.Points.TApiGP.append(newGP)

newGP = factory.TApiGP()
newGP.Lng, newGP.Lat = lat_lng_from_sitecenter(d_west_east/2, -d_south_north/2)
solar_area.Points.TApiGP.append(newGP)

# These should be setup in a more automated way later...

solar_area.Id = 5
solar_area.Description = 'Av godaw'
solar_area.RowAngle = 180 * math.pi / 180
solar_area.TiltAngle = 20 * math.pi / 180
solar_area.InverterSetup = factory.TApiInverterSetup()
solar_area.InverterSetup.MaxPower = 4900  # This is a part of the calculation setup!! It is irelevant here

solar_area.RowSpacing = 10.0
solar_area.PanelCountHor = 2
solar_area.PanelCountVert = 3
solar_area.ShowSubStructure = True
solar_area.AutoRack = True
solar_area.AutoRackData = factory.TApiAutoRackData()
solar_area.AutoRackData.CountX = 3
solar_area.AutoRackData.CountY = 3
solar_area.AutoRackData.Color = factory.TApiColor()
solar_area.AutoRackData.Color.Red = 0
solar_area.AutoRackData.Color.Green = 0
solar_area.AutoRackData.Color.Blue = 255
solar_area.AutoRackData.Color.Alpha = 0
solar_area.AutoRackData.TubeDiameter = 0.3
solar_area.AutoRackData.Margin = 0.03
solar_area.RackDaeFile = os.path.join(path_pv_dea,
                                      'Solarpanel 196x99cm poly silicon landscape_Bifacial.dae')
solar_area.SolarPanelFile = os.path.join(path_pv_panels,
                                         'EMD-Generic_2021_300W_0.99x1.96Mono_3xBypass.PVPanel')
solar_area.PanelOrientation = 'ApipoLandscape'
newColor = factory.TApiColor()
newColor.Red = 255
newColor.Green = 0
newColor.Blue = 0
newColor.Alpha = 0
solar_area.Color = newColor
solar_area.AreaFillMethod = 'ApiafmCenteredUseSpacing'
solar_area.UseAreaFillBuffer = True
solar_area.AreaFillBuffer = 1.5

solar_area.ExclusionArea = False
solar_area.SubStructureHeight = 1.0
solar_area.SubStructureScaleX = 1.0
solar_area.SubStructureScaleY = 1.0
solar_area.SubStructureScaleZ = 1.0
solar_area.SubStructureTranslateX = 0.0
solar_area.SubStructureTranslateY = 0.0
solar_area.SubStructureTranslateZ = 0.0
solar_area.SubStructureAutoPosition = True
solar_area.BiFacial = False
solar_area.GroundReflection = 0.2
solar_area.TakeReflectionFromCalc = False
solar_area.Tracking = True
solar_area.ManualTilt = True
solar_area.TrackingTiltLimitMin = 0.0
solar_area.TrackingTiltLimitMax = math.pi
solar_area.TrackingAzimuthLimitMin = 0.0
solar_area.TrackingAzimuthLimitMax = math.pi / 2
solar_area.TrackingOptimizeBacktracking = False
solar_area.UseReferencePanelForContinousTrackingCalc = True
solar_area.UseAreaFillSpacing = False
solar_area.AreaFillSpacing = 0.0
solar_area.TrackingAutoTiltType = 'ApiattManual'
solar_area.ReferencePanelSetup = factory.TApiReferencePanelSetup()
solar_area.ReferencePanelSetup.ReferenceSpot = factory.TApiGP()
solar_area.ReferencePanelSetup.ReferenceSpot.Lat = 56.173189
solar_area.ReferencePanelSetup.ReferenceSpot.Lng = 10.730992
solar_area.ReferencePanelSetup.ReferenceSpot.ReferenceSpotAuto = False

nan_to_skipvalue(solar_area)
solar_obj.SolarAreas.TApiSolarArea.append(solar_area)
nan_to_skipvalue(solar_obj)

setObj = obj_solar_pv_service.SetSolarPvObject(solar_obj)

# Adding meteo data needed
meteo_data_type = 'ODSClimate'
online_data_service.PrepareService(dataType=meteo_data_type, lat=lat, lon=lng)
meteo_services = online_data_service.GetServices(dataType=meteo_data_type)
print(meteo_services)
# Selecting a specific dataset and downloading it
meteo_service_id = 'sicerra'
meteo_handles_cerra = online_data_service.DownloadMeteoData(implId=meteo_service_id,
                                                      lat=lat,
                                                      lon=lng,
                                                      maxDist=30000,
                                                      numPoints=1,
                                                      fromYear=2014,
                                                      toYear=2015)

# meteo_service_id = 'siERABasic'
# meteo_handles_era = online_data_service.DownloadMeteoData(implId=meteo_service_id,
#                                                       lat=lat,
#                                                       lon=lng,
#                                                       maxDist=30000,
#                                                       numPoints=1,
#                                                       fromYear=2014,
#                                                       toYear=2015)

# Start with solar calculation
new_calc_handle = calculation_service.CreateEmpty("CalcSolarPV")
calculation_service.OpenCalcForEdit(new_calc_handle)
calc_solar_edited = calc_solar_pv_service.GetSolarPvCalc()
calc_solar_pv_service.CloseStatusWindows()
calculation_service.CloseCalc(False)
calculation_service.Calculate(new_calc_handle)

# calc_solar_edited.SolarMeteoData.SolarRadiationMeteoHandle
# calc_solar_edited.SolarMeteoData.LocalRadiationMeteoHandle
# calc_solar_edited.SolarMeteoData.SolarRadiationMeteoHandle
# calc_solar_edited.SolarMeteoData.SolarRadiationMeteoHandle